﻿using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace vamtaco
{
	public class NippleClippingReducer : MVRScript
	{
        public Dictionary<string, DAZMorph> Morphs;

        private Dictionary<string, JSONStorableFloat> _selectedClothingItemMorphValues;

        private JSONStorableString _selectedClothingItem;

        private DAZCharacterSelector _character;
        private GenerateDAZMorphsControlUI _morphControl;

        private JSONStorableStringChooser _clothingList;
        private UIDynamicPopup _clothingListPopup;

        private JSONStorableStringChooser _noClipclothingList;
        private UIDynamicPopup _noClipclothingListPopup;

        private List<NoClipClothingItem> _noClipClothingItems;

        private JSONStorableBool _restoreNipplesOnUndress;


        private string[] _supportedMorphs = new string[] { "Nipple Diameter", "Nipples Diameter", "Nipple Length",
            "Nipples Large", "Nipple Size", "Nipples Size", "Nipples", "Nipples Apply", "Nipples Depth" };

        private Dictionary<string, JSONStorableFloat> _initialMorphValues;

        private NippleClippingReducerJson _nippleClippingReducerJson;

        public override void Init()
		{
			if (containingAtom.type != "Person")
			{
				SuperController.LogMessage("Nipple Clipping Reducer (Disabled): This plugin is designed to be used on a person atom.");
				enabled = false;
				return;
			}

            _noClipClothingItems = new List<NoClipClothingItem>();

            //get morphs list from file...
            //NOTE: Can't write to file in var??
            //var json = MVR.FileManagementSecure.FileManagerSecure.ReadAllText($"{GetPluginPath(this)}/NippleClippingReducer.Data.json");

            if(!MVR.FileManagementSecure.FileManagerSecure.DirectoryExists($"Custom/Scripts/vamtaco/NippleClippingReducer"))
                MVR.FileManagementSecure.FileManagerSecure.CreateDirectory($"Custom/Scripts/vamtaco/NippleClippingReducer");

            if (!MVR.FileManagementSecure.FileManagerSecure.FileExists($"Custom/Scripts/vamtaco/NippleClippingReducer/NippleClippingReducer.Data.json"))
                MVR.FileManagementSecure.FileManagerSecure.WriteAllText($"Custom/Scripts/vamtaco/NippleClippingReducer/NippleClippingReducer.Data.json", "{\"persistedNoClipClothingItems\": []}");

            var json = MVR.FileManagementSecure.FileManagerSecure.ReadAllText($"Custom/Scripts/vamtaco/NippleClippingReducer/NippleClippingReducer.Data.json");
            _nippleClippingReducerJson = new NippleClippingReducerJson(json, _supportedMorphs);

            var geometry = containingAtom.GetStorableByID("geometry");
            _character = geometry as DAZCharacterSelector;
            _morphControl = _character.morphsControlUI;

            //refresh active clothing in dropdown
            var refreshClothingListButton = CreateButton("Refresh Clothing", false);
            refreshClothingListButton.buttonColor = Color.green;
            refreshClothingListButton.button.onClick.AddListener(() =>
            {
                RefreshClothingList();
            });

            CreateSpacer(true).height = refreshClothingListButton.height;

            //create initial clothing list and ui
            var activeClothingNames = _character.clothingItems.Where(c => c.isActiveAndEnabled).Select(c => c.displayName).ToList();
            _clothingList = new JSONStorableStringChooser("ClothingListPopup", activeClothingNames, "", "Active Clothing");
            RegisterStringChooser(_clothingList);
            _clothingListPopup = CreatePopup(_clothingList, false);

            //add selected clothing item to noclip list
            var addClothingItemButton = CreateButton("Add to No Clip List", false);
            addClothingItemButton.buttonColor = Color.green;
            addClothingItemButton.button.onClick.AddListener(() =>
            {
                AddSelectedClothingItemToNoClip();
            });

            var noClipList = _nippleClippingReducerJson.PersistedNoClipClothingItems.Select(o => o.Key).ToList();
            _noClipclothingList = new JSONStorableStringChooser("ClothingListPopup", noClipList, "", "No Clip Items", SelectNoClipClothingItemCallback);
            RegisterStringChooser(_noClipclothingList);
            _noClipclothingListPopup = CreatePopup(_noClipclothingList, true);

            //removed selected clothing item from noclip list
            var removeClothingItemButton = CreateButton("Remove from No Clip List", true);
            removeClothingItemButton.buttonColor = Color.red;
            removeClothingItemButton.button.onClick.AddListener(() =>
            {
                RemoveSelectedClothingItemFromNoClip();
            });

            Morphs = new Dictionary<string, DAZMorph>();
            _initialMorphValues = new Dictionary<string, JSONStorableFloat>();
            _selectedClothingItemMorphValues = new Dictionary<string, JSONStorableFloat>();

            //gather supported morphs for initial restore state
            foreach (var sm in _supportedMorphs)
            {
                var morph = _morphControl.GetMorphByDisplayName(sm);

                if (morph != null)
                {
                    //add initial restore state and for save to scene
                    Morphs.Add(sm, morph);
                    var val = new JSONStorableFloat($"NCR {sm}", 0f, morph.min, morph.max);
                    RegisterFloat(val);
                    val.val = morph.appliedValue;
                    _initialMorphValues.Add(sm, val);

                    //setup support for UI config of no clip values
                    var morphNoClipValue = new JSONStorableFloat($"{sm}"
                    , morph.appliedValue
                    , SetSupportedMorphValueCallback
                    , morph.min
                    , morph.max);

                    var displayName = morph.displayName;
                    if (displayName == string.Empty)
                        displayName = morph.morphName;

                    _selectedClothingItemMorphValues[sm] = morphNoClipValue;
                }
            }

            //name of selected clothing item
            _selectedClothingItem = new JSONStorableString("SelectedClothingItem", "");
            var text = CreateTextField(_selectedClothingItem);

            CreateSpacer(true).height = _selectedClothingItem.dynamicText.height;

            //not working; hide from UI
            _restoreNipplesOnUndress = new JSONStorableBool("RestoreNipplesOnUndress", true);
            _restoreNipplesOnUndress.valNoCallback = false;
            //var restoreNipsToggle = CreateToggle(_restoreNipplesOnUndress, true);
            //restoreNipsToggle.height = _selectedClothingItem.dynamicText.height;
            //restoreNipsToggle.labelText.text = "Restore Nipples on Undress";

            bool right = false;
            //supported morphs
            foreach (var morph in _supportedMorphs)
            {
                CreateSlider(_selectedClothingItemMorphValues[morph], right);
                right = !right;
            }

            //save selected clothing item morph values
            var saveSelectedClothingItemMorphsValuesButton = CreateButton("Update Clothing Settings", false);
            saveSelectedClothingItemMorphsValuesButton.buttonColor = Color.green;
            saveSelectedClothingItemMorphsValuesButton.button.onClick.AddListener(() =>
            {
                SaveSelectedClothingItemMorphsValues();
            });

            //load supported clothing
            //NOTE:  This is going to be a major issue if people add 1000's of clothing options to it
            //OPTIMIZE: Only load the clothing that the containingAtom has active and make user refresh manually if they add new clothing to the chanracter
            foreach (var item in _nippleClippingReducerJson.PersistedNoClipClothingItems)
            {
                LoadNoClipClothingItem(item.Key);
            }
        }

        void Start()
        {
            //if any clothing is active, reduce the nipples
            if (_noClipClothingItems.Any(o => o.ClothingItem.isActiveAndEnabled))
            {
                _noClipclothingListPopup.popup.currentValue = _noClipClothingItems.FirstOrDefault(o => o.ClothingItem.isActiveAndEnabled)?.Name;
                ReduceNippleSize();
            }
        }

        void Update()
        {
            foreach(var c in _noClipClothingItems)
            {
                c.EvaluateStateChange();
            }
        }

        private void AddSelectedClothingItemToNoClip()
        {
            if(!_noClipClothingItems.Any(c => c.Name == _clothingListPopup.popup.currentValue))
            {
                var clothing = _character.clothingItems.SingleOrDefault(c => c.displayName == _clothingListPopup.popup.currentValue);
                if (clothing != null)
                {
                    var ncci = new NoClipClothingItem(clothing);
                    ncci.OnStateChanged += new NoClipClothingItem.OnStateChangedEvent(OnClothingItemStateChanged);
                    _noClipClothingItems.Add(ncci);

                    if (clothing.isActiveAndEnabled)
                        ReduceNippleSize();

                    var pncci = new PersistedNoClipClothingItem { Name = clothing.displayName };
                    foreach (var sm in _supportedMorphs)
                    {
                        pncci.MorphValues.Add(sm, new NoClipMorphValue { MorphName = sm, Value = Morphs[sm].min });
                    }
                    _nippleClippingReducerJson.PersistedNoClipClothingItems.Add(clothing.displayName, pncci);

                    SavePluginData();

                    var list = _nippleClippingReducerJson.PersistedNoClipClothingItems.Select(o => o.Key).ToList();
                    list.Insert(0, "");
                    _noClipclothingList.choices = list;
                    _noClipclothingList.popup.currentValue = clothing.displayName;
                    ShowSelectedClothingItemUI();
                }
            }
        }

        private void RemoveSelectedClothingItemFromNoClip()
        {
            string selected = _noClipclothingListPopup.popup.currentValueNoCallback;
            _nippleClippingReducerJson.PersistedNoClipClothingItems.Remove(selected);

            if (_noClipClothingItems.Any(c => c.Name == selected))
            {
                var ncci = _noClipClothingItems.Single(c => c.Name == selected);

                if (ncci.ClothingItem.isActiveAndEnabled)
                    RestoreNippleSize();

                _noClipClothingItems.Remove(ncci);
            }

            SavePluginData();

            _noClipclothingList.choices = _nippleClippingReducerJson.PersistedNoClipClothingItems.Select(o => o.Key).ToList();
            _noClipclothingListPopup.popup.currentValue = "";
        }

        public void OnClothingItemStateChanged(DAZClothingItem clothingItem)
        {
            //item is now active so reduce nipple size
            if (clothingItem.isActiveAndEnabled)
            {
                var clothSimControlChildren = clothingItem.GetComponentsInChildren<ClothSimControl>();
                bool anyUndressed = clothSimControlChildren.Any(o => o.clothSettings.BreakEnabled);

                if (anyUndressed && _restoreNipplesOnUndress.valNoCallback)
                {
                    RestoreNippleSize();
                    return;
                }

                _noClipclothingListPopup.popup.currentValue = clothingItem.displayName;
                ReduceNippleSize();
                return;
            }

            _noClipclothingListPopup.popup.currentValue = string.Empty;
            RestoreNippleSize();
        }

        private void ReduceNippleSize()
        {
            var selected = _noClipclothingListPopup.popup.currentValueNoCallback;

            //make all nipple morphs smaller to prevent poke through
            foreach (var morph in Morphs)
            {
                if(_nippleClippingReducerJson.PersistedNoClipClothingItems.ContainsKey(selected) &&
                    _nippleClippingReducerJson.PersistedNoClipClothingItems[selected].MorphValues.ContainsKey(morph.Key))
                    morph.Value.morphValue = _nippleClippingReducerJson.PersistedNoClipClothingItems[selected].MorphValues[morph.Key].Value;
                else
                    morph.Value.morphValue = morph.Value.min;
            }
        }

        private void RefreshClothingList()
        {
            _clothingList.choices.Clear();
            _clothingList.choices = _character.clothingItems.Where(c => c.isActiveAndEnabled).Select(c => c.displayName).ToList();
            _clothingList.SetValToDefault();
        }

        private void RestoreNippleSize()
        {
            //restore nipples to original value
            foreach (var morph in Morphs)
            {
                morph.Value.morphValue = _initialMorphValues[morph.Key].val;
            }
        }

        private void LoadNoClipClothingItem(string name)
        {
            var clothing = _character.clothingItems.SingleOrDefault(c => c.displayName == name);
            if (clothing != null)
            {
                var ncci = new NoClipClothingItem(clothing);
                ncci.OnStateChanged += new NoClipClothingItem.OnStateChangedEvent(OnClothingItemStateChanged);
                _noClipClothingItems.Add(ncci);
            }
        }

        private void SavePluginData()
        {
            var json = _nippleClippingReducerJson.ToJSON();
            //NOTE: Can't save to file in var???
            //MVR.FileManagementSecure.FileManagerSecure.WriteAllText($"{GetPluginPath(this)}/NippleClippingReducer.Data.json", json);
            MVR.FileManagementSecure.FileManagerSecure.WriteAllText($"Custom/Scripts/vamtaco/NippleClippingReducer/NippleClippingReducer.Data.json", json);
        }

        private void SaveSelectedClothingItemMorphsValues()
        {
            if (_noClipclothingListPopup.popup.currentValueNoCallback == string.Empty)
                return;

            if (_nippleClippingReducerJson.PersistedNoClipClothingItems.ContainsKey(_noClipclothingListPopup.popup.currentValueNoCallback))
            {
                var persisted = _nippleClippingReducerJson.PersistedNoClipClothingItems[_noClipclothingListPopup.popup.currentValueNoCallback];
                persisted.RestoreNipplesOnUndress = _restoreNipplesOnUndress.valNoCallback;

                foreach (var sm in _supportedMorphs)
                {
                    if (!persisted.MorphValues.ContainsKey(sm))
                    {
                        persisted.MorphValues.Add(sm, new NoClipMorphValue { MorphName = sm });
                    }

                    persisted.MorphValues[sm].Value = _selectedClothingItemMorphValues[sm].valNoCallback;
                }
                SavePluginData();
            }
        }

        private void SelectNoClipClothingItemCallback(string value)
        {
            ShowSelectedClothingItemUI();
        }

        private void ShowSelectedClothingItemUI()
        {
            string selected = _noClipclothingListPopup.popup.currentValue;
            if (selected == string.Empty)
            {
                HideSelectedClothingItemUI();
                return;
            }

            _selectedClothingItem.val = $"Editing: {selected}";
            _selectedClothingItem.dynamicText.enabled = true;
            _restoreNipplesOnUndress.valNoCallback = _nippleClippingReducerJson.PersistedNoClipClothingItems[selected].RestoreNipplesOnUndress;

            foreach (var persisted in _selectedClothingItemMorphValues)
            {
                if (_nippleClippingReducerJson.PersistedNoClipClothingItems[selected].MorphValues.ContainsKey(persisted.Key))
                    persisted.Value.val = _nippleClippingReducerJson.PersistedNoClipClothingItems[selected].MorphValues[persisted.Key].Value;
                else
                    persisted.Value.val = persisted.Value.defaultVal;

                persisted.Value.slider.enabled = true;
            }
        }

        private void HideSelectedClothingItemUI()
        {
            _selectedClothingItem.val = "Editing: None Selected";
            //reset all to default with no callback
        }

        private void ClearExistingSelectedUI()
        {
            _selectedClothingItem.text = null;
            foreach (var sm in _selectedClothingItemMorphValues)
            {
                sm.Value.slider = null;
            }
        }

        private void SetSupportedMorphValueCallback(float value)
        {
            if (_noClipclothingListPopup.popup.currentValueNoCallback == string.Empty)
                return;

            foreach (var morph in Morphs)
            {
                morph.Value.morphValue = _selectedClothingItemMorphValues[morph.Key].val;
            }
        }

        // u/MacGruber_VR: https://www.patreon.com/MacGruber_Laboratory/posts
        // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
        public static string GetPluginPath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }
    }

    public class NoClipClothingItem
    {
        public string Name { get { return ClothingItem.displayName; } }
        
        public DAZClothingItem ClothingItem { get; }

        public delegate void OnStateChangedEvent(DAZClothingItem clothingItem);
        public OnStateChangedEvent OnStateChanged;

        private bool _previousState;
        private bool _previousUndressState;
        private ClothSimControl[] _clothSimControls;

        public NoClipClothingItem(DAZClothingItem clothingItem)
        {
            ClothingItem = clothingItem;
            _previousState = clothingItem.isActiveAndEnabled;
            _clothSimControls = clothingItem.GetComponentsInChildren<ClothSimControl>();
            _previousUndressState = _clothSimControls.Any(o => o.clothSettings.BreakEnabled);
        }

        public void EvaluateStateChange()
        {
            if (ClothingItem == null)
                return;

            //clothing item state has changed
            bool anyUndressed = _clothSimControls.Any(o => o.clothSettings.BreakEnabled);

            if (ClothingItem.isActiveAndEnabled != _previousState
                || anyUndressed != _previousUndressState)
            {
                _previousState = ClothingItem.isActiveAndEnabled;
                _previousUndressState = anyUndressed;
                OnStateChanged?.Invoke(ClothingItem);
            }
        }
    }

    public class NippleClippingReducerJson
    {
        public Dictionary<string, PersistedNoClipClothingItem> PersistedNoClipClothingItems;

        public NippleClippingReducerJson(string json, string[] supportedMorphs)
        {
            PersistedNoClipClothingItems = new Dictionary<string, PersistedNoClipClothingItem>();

            var node = SimpleJSON.JSON.Parse(json);
            var array = node["persistedNoClipClothingItems"].AsArray;
            for (int i = 0; i < array.Count; i++)
            {
                var name = array[i]["name"].Value;

                var add = new PersistedNoClipClothingItem();
                add.Name = name;
                add.RestoreNipplesOnUndress = array[i]["restoreNipplesOnUndress"].AsBool;

                foreach (var sm in supportedMorphs)
                {
                    var m = array[i][sm];
                    if (sm != null)
                        add.MorphValues.Add(sm, new NoClipMorphValue { MorphName = sm, Value = m.AsFloat });
                }

                PersistedNoClipClothingItems.Add(add.Name, add);
            }
        }

        public string ToJSON()
        {
            var json = new JSONClass();
            var array = new JSONArray();
            int index = 0;

            foreach (var clothing in PersistedNoClipClothingItems)
            {
                array[index]["name"] = clothing.Key;
                array[index]["restoreNipplesOnUndress"] = new JSONData(clothing.Value.RestoreNipplesOnUndress);

                foreach (var mv in clothing.Value.MorphValues)
                {
                    array[index][mv.Key] = new JSONData(mv.Value.Value);
                }

                index++;
            }
            
            json.Add("persistedNoClipClothingItems", array);
            return json.ToString();
        }   
    }

    public class PersistedNoClipClothingItem
    {
        public string Name { get; set; }
        public bool RestoreNipplesOnUndress { get; set; }

        public Dictionary<string, NoClipMorphValue> MorphValues { get; set; }

        public PersistedNoClipClothingItem()
        {
            RestoreNipplesOnUndress = false;
            MorphValues = new Dictionary<string, NoClipMorphValue>();
        }
    }

    public class NoClipMorphValue
    {
        public string MorphName { get; set; }
        public float Value { get; set; }
    }
}
